from datetime import datetime

import pytest

from nowcasting_datamodel.fake import make_fake_pv_system
from nowcasting_datamodel.models import PVSystem, PVSystemSQL, PVYield


def test_make_fake_pv_system():
    pv_system_sql: PVSystemSQL = make_fake_pv_system()
    pv_system = PVSystem.model_validate(pv_system_sql)
    _ = PVSystem.to_orm(pv_system)


def test_pv_system_error():
    with pytest.raises(Exception):
        _ = PVSystem(pv_system_id=1, provider="fake.com", latitude=55, longitude=0)


def test_make_fake_pv_yield():
    pv_yield = PVYield(datetime_utc=datetime(2022, 1, 1), solar_generation_kw=-1)
    assert pv_yield.solar_generation_kw == 0

    _ = pv_yield.to_orm()



"""Read database functions

1. Get the one metric
2. get datetime interval
"""

import logging
from datetime import datetime, timezone
from typing import List, Optional

from sqlalchemy.orm.session import Session

from nowcasting_datamodel.models import DatetimeIntervalSQL, MetricSQL, MetricValueSQL, MLModelSQL

logger = logging.getLogger(__name__)


def get_metric(session: Session, name: str) -> MetricSQL:
    """
    Get metric object from database

    :param session: database session
    :param name: name of the metric

    return: One Metric SQl object

    """

    # start main query
    query = session.query(MetricSQL)

    # filter on name
    query = query.filter(MetricSQL.name == name)

    # get all results
    metrics = query.all()

    if len(metrics) == 0:
        logger.debug(f"Metric for name {name} does not exist so going to add it")

        metric = MetricSQL(name=name, description="TODO need to update")

        session.add(metric)
        session.commit()

    else:
        metric = metrics[0]

    return metric


def get_datetime_interval(
    session: Session, start_datetime_utc: datetime, end_datetime_utc: datetime
) -> DatetimeIntervalSQL:
    """
    Get datetime interval object from database

    :param session: database session
    :param start_datetime_utc: start datetime of the interval
    :param end_datetime_utc: end datetime of the interval

    return: One Datetime interval SQl object

    """

    # start main query
    query = session.query(DatetimeIntervalSQL)

    # filter on start_datetime_utc
    query = query.filter(DatetimeIntervalSQL.start_datetime_utc == start_datetime_utc)

    # filter on end_datetime_utc
    query = query.filter(DatetimeIntervalSQL.end_datetime_utc == end_datetime_utc)

    # get all results
    datetime_intervals = query.all()

    if len(datetime_intervals) == 0:
        logger.debug(
            f"Datetime interbal for {start_datetime_utc=} "
            f"and {end_datetime_utc=} does not exist so going to add it"
        )

        datetime_interval = DatetimeIntervalSQL(
            start_datetime_utc=start_datetime_utc, end_datetime_utc=end_datetime_utc
        )

        session.add(datetime_interval)
        session.commit()

    else:
        datetime_interval = datetime_intervals[0]

    return datetime_interval


def read_latest_me_national(
    session: Session, metric_name: str = "Half Hourly ME", model_name: Optional[str] = None
) -> List[MetricValueSQL]:
    """
    Get the latest me for the national forecast

    This is grouped by 'time_of_day' and 'forecast_horizon_minutes'

    :param session: database sessions
    :param metric_name: metric name, defaulted to "Half Hourly ME"
    :param model_name: model name, defaulted to None
    :return: list of MetricValueSQL for ME for each 'time_of_day' and 'forecast_horizon_minutes'
    """

    logger.debug(f"Reading latest {metric_name=} for {model_name=}")

    # start main query
    query = session.query(MetricValueSQL)
    query = query.join(MetricSQL)

    query = query.distinct(MetricValueSQL.time_of_day, MetricValueSQL.forecast_horizon_minutes)

    # filter on metric name
    query = query.filter(MetricSQL.name == metric_name)

    query = query.order_by(
        MetricValueSQL.time_of_day,
        MetricValueSQL.forecast_horizon_minutes,
        MetricValueSQL.created_utc.desc(),
    )

    if model_name is not None:
        query = query.join(MLModelSQL)
        query = query.filter(MLModelSQL.name == model_name)

    # get all results
    metric_values = query.all()

    # add timezone, show be the same datetime interval for all
    if len(metric_values) > 0:
        datetime_interval = metric_values[0].datetime_interval
        datetime_interval.start_datetime_utc = datetime_interval.start_datetime_utc.replace(
            tzinfo=timezone.utc
        )
        datetime_interval.end_datetime_utc = datetime_interval.end_datetime_utc.replace(
            tzinfo=timezone.utc
        )

    return metric_values



"""Utils functions for test"""

from datetime import datetime, timezone

# Used constants
from nowcasting_datamodel.utils import convert_to_camelcase, datetime_with_timezone


def test_datetime_with_timezone_handles_none():
    assert datetime_with_timezone(None, None) is None


def test_datetime_with_timezone_handles_string():
    dt = datetime_with_timezone(None, "2024-07-14T18:01:26+00:00")
    assert isinstance(dt, datetime)
    assert dt.tzinfo == timezone.utc


def test_datetime_with_timezone():
    """Test function datetime_with_timezone"""

    time_now = datetime.now(timezone.utc)

    # check functions works
    new_time_now = datetime_with_timezone(None, time_now)
    assert new_time_now == time_now

    # check functions adds timezone
    time_now = datetime.now()
    time_now = datetime_with_timezone(None, time_now)
    assert new_time_now != time_now
    assert new_time_now.tzinfo == timezone.utc


def test_convert_to_camelcase():
    """Test convert to camelcase works"""
    assert convert_to_camelcase("foo_bar") == "fooBar"
    assert convert_to_camelcase("foo_bar_baz") == "fooBarBaz"



"""Read user"""

import logging
from datetime import datetime
from typing import List, Optional

from sqlalchemy.orm import contains_eager
from sqlalchemy.orm.session import Session

from nowcasting_datamodel.models.api import APIRequestSQL, UserSQL

logger = logging.getLogger(__name__)


def get_user(session: Session, email: str) -> UserSQL:
    """
    Get metric object from database

    :param session: database session
    :param email: email of user
    return: One Metric SQl object
    """

    # start main query
    query = session.query(UserSQL)
    # filter on name
    query = query.filter(UserSQL.email == email)
    # get all results
    users = query.all()
    if len(users) == 0:
        logger.debug(f"User for name {email} does not exist so going to add it")
        user = UserSQL(email=email)
        session.add(user)
        session.commit()
    else:
        user = users[0]
    return user


def get_all_last_api_request(
    session: Session,
    include_in_url: Optional[str] = None,
    exclude_in_url: Optional[str] = None,
    start_datetime: Optional[datetime] = None,
    end_datetime: Optional[datetime] = None,
) -> List[APIRequestSQL]:
    """
    Get all last api requests for all users

    :param session: database session
    :param include_in_url: Optional filter to include only URLs containing this string
    :param exclude_in_url: Optional filter to exclude URLs containing this string
    :param start_datetime: only get api requests after start datetime
    :param end_datetime: only get api requests before end datetime
    :return: List of last API requests
    """

    query = (
        session.query(APIRequestSQL)
        .distinct(APIRequestSQL.user_uuid)
        .join(UserSQL)
        .options(contains_eager(APIRequestSQL.user))
        .populate_existing()
        .order_by(APIRequestSQL.user_uuid, APIRequestSQL.created_utc.desc())
    )
    # Apply URL filtering if specified
    if include_in_url is not None:
        query = query.filter(APIRequestSQL.url.like(f"%{include_in_url}%"))

    if exclude_in_url is not None:
        query = query.filter(~APIRequestSQL.url.like(f"%{exclude_in_url}%"))

    if start_datetime is not None:
        query = query.filter(APIRequestSQL.created_utc >= start_datetime)

    if end_datetime is not None:
        query = query.filter(APIRequestSQL.created_utc <= end_datetime)

    return query.all()


def get_api_requests_for_one_user(
    session: Session,
    email: str,
    start_datetime: Optional[datetime] = None,
    end_datetime: Optional[datetime] = None,
    include_in_url: Optional[str] = None,
    exclude_in_url: Optional[str] = None,
) -> List[APIRequestSQL]:
    """
    Get all api requests for one user

    :param session: database session
    :param email: user email
    :param start_datetime: only get api requests after start datetime
    :param end_datetime: only get api requests before end datetime
    :param include_in_url: Optional filter to include only URLs containing this string
    :param exclude_in_url: Optional filter to exclude URLs containing this string
    """

    query = session.query(APIRequestSQL).join(UserSQL).filter(UserSQL.email == email)

    if start_datetime is not None:
        query = query.filter(APIRequestSQL.created_utc >= start_datetime)

    if end_datetime is not None:
        query = query.filter(APIRequestSQL.created_utc <= end_datetime)

    # Apply URL filtering if specified
    if include_in_url is not None:
        query = query.filter(APIRequestSQL.url.like(f"%{include_in_url}%"))

    if exclude_in_url is not None:
        query = query.filter(~APIRequestSQL.url.like(f"%{exclude_in_url}%"))

    return query.order_by(APIRequestSQL.created_utc.desc()).all()



from typing import List

from nowcasting_datamodel.fake import N_FAKE_FORECASTS
from nowcasting_datamodel.models import ForecastSQL


def test_get_session(db_connection):
    with db_connection.get_session() as _:
        pass


def test_read_forecast(db_session):
    forecasts = db_session.query(ForecastSQL).all()
    assert len(forecasts) == 0


def test_read_forecast_one(db_session, forecast_sql):
    # query
    forecasts: List[ForecastSQL] = db_session.query(ForecastSQL).all()
    assert len(forecasts) == 1
    assert forecast_sql[0] == forecasts[0]
    assert len(forecasts[0].forecast_values) == N_FAKE_FORECASTS



"""App for running database migrations"""

import logging
import os
import time

import click
from alembic import command
from alembic.config import Config

filename = os.path.dirname(os.path.abspath(__file__)) + "/alembic.ini"

logger = logging.getLogger(__name__)
logger.setLevel(os.getenv("LOGLEVEL", "INFO"))


@click.command()
@click.option(
    "--make-migrations",
    default=False,
    envvar="MAKE_MIGRATIONS",
    help="Option to make database migrations",
    type=click.BOOL,
    is_flag=True,
)
@click.option(
    "--run-migrations",
    default=False,
    envvar="RUN_MIGRATIONS",
    help="Option to run the database migrations",
    type=click.BOOL,
    is_flag=True,
)
def app(make_migrations: bool, run_migrations: bool):
    """
    Make migrations and run them

    :param make_migrations: option to make the database migrations
    :param run_migrations: option to run migrations
    :param run_pv: option to run PV database as well
    """

    # some times need 1 second for the database to get started
    time.sleep(1)

    if make_migrations:
        make_all_migrations("forecast")
        # make_all_migrations("pv")

    if run_migrations:
        run_all_migrations("forecast")
        # run_all_migrations("pv")


def make_all_migrations(database: str):
    """
    Make migrations.

    They are saved in
    nowcasting_datamodel/alembic/{database}/version

    :param database: either 'pv' or 'forecast'
    """

    logger.info(f"Making migrations for {database}")

    alembic_cfg = Config(filename, ini_section=database)
    command.revision(alembic_cfg, autogenerate=True)

    logger.info("Making migrations:done")


def run_all_migrations(database: str):
    """Run migrations

    Looks at the migrations in
    nowcasting_datamodel/alembic/{database}/version
    and runs them
    """

    logger.info(f"Running migrations for {database}")

    # see here - https://alembic.sqlalchemy.org/en/latest/api/config.html
    # for more help
    alembic_cfg = Config(filename, ini_section=database)

    command.upgrade(alembic_cfg, "head")

    logger.info(f"Running migrations {database}: done")


if __name__ == "__main__":
    app()



"""Test national forecast"""

# Used constants
import pytest

from nowcasting_datamodel import N_GSP
from nowcasting_datamodel.models import LocationSQL
from nowcasting_datamodel.models.forecast import Forecast, ForecastSQL
from nowcasting_datamodel.national import make_national_forecast


def test_make_national_forecast(forecasts_all, db_session):
    locations = db_session.query(LocationSQL).all()
    assert len(locations) == N_GSP

    national_forecast = make_national_forecast(forecasts=forecasts_all, session=db_session)
    db_session.add(national_forecast)
    db_session.commit()

    assert type(national_forecast) == ForecastSQL

    locations = db_session.query(LocationSQL).all()
    assert len(locations) == N_GSP + 1

    # run it again, check there are still 339 locations
    national_forecast = make_national_forecast(forecasts=forecasts_all, session=db_session)
    db_session.add(national_forecast)
    db_session.commit()
    locations = db_session.query(LocationSQL).all()
    assert len(locations) == N_GSP + 1


def test_make_national_forecast_error(forecasts_all, db_session):
    forecasts_all = [Forecast.model_validate(f, from_attributes=True) for f in forecasts_all]
    forecasts_all[0].location.gsp_id = 2

    with pytest.raises(Exception):
        _ = make_national_forecast(forecasts=forecasts_all, session=db_session)



"""Pydantic and Sqlalchemy models for the database

The following class are made
1. Reusable classes
3. Model object, what forecast model is used
4. ForecastValue objects, specific values of a forecast and time
5. Input data status, shows when the data was collected
6. Forecasts, a forecast that is made for one gsp, for several time steps into the future

"""

import logging
from datetime import datetime
from typing import Optional

from pydantic import Field, field_validator
from sqlalchemy import Column, DateTime, Index, Integer, String
from sqlalchemy.orm import relationship

from nowcasting_datamodel.models.base import Base_Forecast
from nowcasting_datamodel.models.utils import CreatedMixin, EnhancedBaseModel
from nowcasting_datamodel.utils import datetime_with_timezone

national_gb_label = "National-GB"

logger = logging.getLogger(__name__)
# TODO #3 Add forecast latest table, this make it easy to load the latest forecast


########
# 3. Model
########
class MLModelSQL(Base_Forecast, CreatedMixin):
    """ML model that is being used"""

    __tablename__ = "model"

    id = Column(Integer, primary_key=True)
    name = Column(String)
    version = Column(String)

    forecast = relationship("ForecastSQL", back_populates="model")
    metric_value = relationship("MetricValueSQL", back_populates="model")


class MLModel(EnhancedBaseModel):
    """ML model that is being used"""

    name: Optional[str] = Field(..., description="The name of the model", index=True)
    version: Optional[str] = Field(..., description="The version of the model")

    def to_orm(self) -> MLModelSQL:
        """Change model to MLModelSQL"""
        return MLModelSQL(
            name=self.name,
            version=self.version,
        )


########
# 4. ForecastValue
# A. SQL table saves all forecast values
# B. SQL table that save the latest forecast
# C. Pydantic object
########


########
# 5. Input data status
########
class InputDataLastUpdatedSQL(Base_Forecast, CreatedMixin):
    """Information about the input data that was used to create the forecast"""

    __tablename__ = "input_data_last_updated"

    id = Column(Integer, primary_key=True)
    gsp = Column(DateTime(timezone=True))
    nwp = Column(DateTime(timezone=True))
    pv = Column(DateTime(timezone=True))
    satellite = Column(DateTime(timezone=True))

    forecast = relationship("ForecastSQL", back_populates="input_data_last_updated")

    Index("index_input_data", CreatedMixin.created_utc.desc())


class InputDataLastUpdated(EnhancedBaseModel):
    """Information about the input data that was used to create the forecast"""

    gsp: datetime = Field(..., description="The time when the input GSP data was last updated")
    nwp: datetime = Field(..., description="The time when the input NWP data was last updated")
    pv: datetime = Field(..., description="The time when the input PV data was last updated")
    satellite: datetime = Field(
        ..., description="The time when the input satellite data was last updated"
    )

    @field_validator("gsp", mode="before")
    def normalize_gsp(cls, v):
        """Normalize gsp field"""
        return datetime_with_timezone(cls, v)

    @field_validator("nwp", mode="before")
    def normalize_nwp(cls, v):
        """Normalize nwp field"""
        return datetime_with_timezone(cls, v)

    @field_validator("pv", mode="before")
    def normalize_pv(cls, v):
        """Normalize pv field"""
        return datetime_with_timezone(cls, v)

    @field_validator("satellite", mode="before")
    def normalize_satellite(cls, v):
        """Normalize satellite field"""
        return datetime_with_timezone(cls, v)

    def to_orm(self) -> InputDataLastUpdatedSQL:
        """Change model to InputDataLastUpdatedSQL"""
        return InputDataLastUpdatedSQL(
            gsp=self.gsp,
            nwp=self.nwp,
            pv=self.pv,
            satellite=self.satellite,
        )


########
# 7. Status
########


class StatusSQL(Base_Forecast, CreatedMixin):
    """Status SQL Model"""

    __tablename__ = "status"

    id = Column(Integer, primary_key=True)
    status = Column(String)
    message = Column(String)


class Status(EnhancedBaseModel):
    """Status Model for a single message"""

    status: str = Field(..., description="Status description")
    message: str = Field(..., description="Status Message")

    def to_orm(self) -> StatusSQL:
        """Change model to SQL"""
        return StatusSQL(status=self.status, message=self.message)

    @field_validator("status")
    def validate_solar_generation_kw(cls, v):
        """Validate the solar_generation_kw field"""
        if v not in ["ok", "warning", "error"]:
            message = f"Status is {v}, not one of 'ok', 'warning', 'error'"
            logger.debug(message)
            raise Exception(message)
        return v



"""Utils functions for models"""

import logging
from datetime import datetime, timezone
from typing import Any, Optional

logger = logging.getLogger(__name__)


def datetime_with_timezone(cls, v: Any) -> Optional[datetime]:
    """Convert to a datetime with a timezone"""
    dt = v
    if not isinstance(dt, datetime):
        if dt is None:
            return dt
        elif isinstance(dt, str):
            dt = datetime.fromisoformat(dt)
        else:
            raise TypeError(f"argument must be a datetime or a str, but was {dt.type()}")
    return _datetime_must_have_timezone(cls, dt)


def _datetime_must_have_timezone(cls, v: datetime) -> Optional[datetime]:
    """Enforce that this variable must have a timezone"""
    if v is None:
        return v
    if v.tzinfo is None:
        logger.debug(f"{v} must have a timezone, for cls {cls}." f"So we are going to add UTC ")
        v = v.replace(tzinfo=timezone.utc)
    return v


def convert_to_camelcase(snake_str: str) -> str:
    """Converts a given snake_case string into camelCase"""
    first, *others = snake_str.split("_")
    return "".join([first.lower(), *map(str.title, others)])



"""
The followin tables are made with sqlamyc and pydantic

1. metric, stores the metric metadata
2. datetime_interval - The datetime interval that the metric is for
3. metric_value, the actual value of this metric

"""

from datetime import datetime, time
from typing import ClassVar, Optional

from pydantic import Field, field_validator
from sqlalchemy import Boolean, Column, DateTime, Float, ForeignKey, Index, Integer, String, Time
from sqlalchemy.orm import relationship

from nowcasting_datamodel.models.base import Base_Forecast
from nowcasting_datamodel.models.gsp import Location
from nowcasting_datamodel.models.utils import CreatedMixin, EnhancedBaseModel
from nowcasting_datamodel.utils import datetime_with_timezone

########
# 1. Metric
########


class MetricSQL(Base_Forecast):
    """Metric metadata"""

    __tablename__ = "metric"

    id = Column(Integer, primary_key=True)
    name = Column(String)
    description = Column(String)

    metric_value = relationship("MetricValueSQL", back_populates="metric")


class Metric(EnhancedBaseModel):
    """Metric metadata"""

    name: str = Field(..., description="The name of the metric")
    description: str = Field(..., description="The description of the metric")

    rm_mode: ClassVar[bool] = True

    def to_orm(self) -> MetricSQL:
        """Change model to LocationSQL"""

        if isinstance(self, MetricSQL):
            return self

        return MetricSQL(
            name=self.name,
            description=self.description,
        )


########
# 2. Datetime Interval
########


class DatetimeIntervalSQL(Base_Forecast):
    """Datetime interval data"""

    __tablename__ = "datetime_interval"

    # add a unique condition on 'gsp_id' and 'target_time'
    __table_args__ = (
        Index(
            "uix_2",  # Index name
            "start_datetime_utc",
            "end_datetime_utc",  # Columns which are part of the index
            unique=True,
            postgresql_where=Column("is_primary"),  # The condition
        ),
    )

    id = Column(Integer, primary_key=True)
    start_datetime_utc = Column(DateTime, index=True)
    end_datetime_utc = Column(DateTime, index=True)
    elexon_settlement_period = Column(Integer, nullable=True)
    is_primary = Column(Boolean, default=True)

    metric_value = relationship("MetricValueSQL", back_populates="datetime_interval")

    Index("ix_start_datetime_utc", start_datetime_utc.desc())
    Index("ix_end_datetime_utc", end_datetime_utc.desc())

    # add unique on start and end datetime


class DatetimeInterval(EnhancedBaseModel):
    """Datetime interval data"""

    start_datetime_utc: datetime = Field(
        ..., description="The timestamp of the start of the interval"
    )
    end_datetime_utc: datetime = Field(..., description="The timestamp of the end of the interval")
    elexon_settlement_period: Optional[datetime] = Field(
        None, description="The elexon settlement period. This is optional"
    )

    @field_validator("start_datetime_utc", mode="before")
    def normalize_start_datetime_utc(cls, v):
        """Normalize start_datetime_utc field"""
        return datetime_with_timezone(cls, v)

    @field_validator("end_datetime_utc", mode="before")
    def normalize_end_datetime_utc(cls, v):
        """Normalize end_datetime_utc field"""
        return datetime_with_timezone(cls, v)

    def to_orm(self) -> DatetimeIntervalSQL:
        """Change model to GSPYieldSQL"""
        return DatetimeIntervalSQL(
            start_datetime_utc=self.start_datetime_utc,
            end_datetime_utc=self.end_datetime_utc,
            elexon_settlement_period=self.elexon_settlement_period,
        )


########
# 3. Metirc Value
########


class MetricValueSQL(Base_Forecast, CreatedMixin):
    """The calculated metric value"""

    __tablename__ = "metric_value"

    id = Column(Integer, primary_key=True)
    value = Column(Float, nullable=False)
    number_of_data_points = Column(Integer, nullable=False)
    forecast_horizon_minutes = Column(Integer, nullable=True)
    time_of_day = Column(Time, nullable=True)
    model_name = Column(String, nullable=True)
    p_level = Column(Float, nullable=True)

    # many (metric values) to one (metric)
    metric = relationship("MetricSQL", back_populates="metric_value")
    metric_id = Column(Integer, ForeignKey("metric.id"), index=True)

    # many (metric values) to one (location)
    location = relationship("LocationSQL", back_populates="metric_value")
    location_id = Column(Integer, ForeignKey("location.id"), index=True)

    # many (metric values) to one (location)
    datetime_interval = relationship("DatetimeIntervalSQL", back_populates="metric_value")
    datetime_interval_id = Column(Integer, ForeignKey("datetime_interval.id"), index=True)

    # many (metric values) to one (model)
    model = relationship("MLModelSQL", back_populates="metric_value")
    model_id = Column(Integer, ForeignKey("model.id"), index=True)


class MetricValue(EnhancedBaseModel):
    """Location that the forecast is for"""

    value: float = Field(..., description="The value of the metric")
    number_of_data_points: int = Field(
        ..., description="The number of data points used to make the value"
    )
    forecast_horizon_minutes: int = Field(
        None,
        description="The forecast horizon in minutes. "
        "60 minutes means the forecast made 60 mintues before the target time",
    )
    time_of_day: Optional[time] = Field(None, description="the time of tday for this metric")
    metric: Metric = Field(..., description="The metric this value is about")
    datetime_interval: DatetimeInterval = Field(
        ..., description="The datetime interval this value is about"
    )
    location: Location = Field(None, description="The location object for this metric value")

    rm_mode: ClassVar[bool] = True

    def to_orm(self) -> MetricValueSQL:
        """Change model to MetricValueSQL"""

        if isinstance(self, MetricValueSQL):
            return self

        return MetricValueSQL(
            value=self.value,
            number_of_data_points=self.number_of_data_points,
            metric=self.metric.to_orm(),
            location=self.location.to_orm(),
            datetime_interval=self.datetime_interval.to_orm(),
            forecast_horizon_minutes=self.forecast_horizon_minutes,
            time_of_day=self.time_of_day,
        )



"""Read functions for models"""

from datetime import datetime
from typing import Optional

from sqlalchemy.orm import Session

from nowcasting_datamodel.models import MLModelSQL
from nowcasting_datamodel.models.forecast import ForecastSQL
from nowcasting_datamodel.read.read import logger


def get_models(
    session, with_forecasts: Optional[bool] = False, forecast_created_utc: Optional[datetime] = None
) -> MLModelSQL:
    """
    Get all models from the database, distinct on name

    Can also filter if the model has forecasts, and if they are made after a certain time

    :param session: sql session
    :param with_forecasts: only get models with forecasts
    :param forecast_created_utc: only look at forecast created after this time
    :return:
    """

    query = session.query(MLModelSQL)
    query = query.distinct(MLModelSQL.name)

    if with_forecasts:
        query = query.join(ForecastSQL)

        if forecast_created_utc:
            query = query.filter(ForecastSQL.created_utc > forecast_created_utc)

    # order by
    query = query.order_by(MLModelSQL.name, MLModelSQL.created_utc.desc())

    # get all results
    models = query.all()

    return models


def get_model(session: Session, name: str, version: Optional[str] = None) -> MLModelSQL:
    """
    Get model object from name and version

    :param session: database session
    :param name: name of the model
    :param version: version of the model

    return: Model object

    """

    # start main query
    query = session.query(MLModelSQL)

    # filter on gsp_id
    query = query.filter(MLModelSQL.name == name)
    if version is not None:
        query = query.filter(MLModelSQL.version == version)

    # gets the latest version
    query = query.order_by(MLModelSQL.created_utc.desc())

    # get all results
    models = query.all()

    if len(models) == 0:
        logger.debug(
            f"Model for name {name} and version {version} does not exist so going to add it"
        )

        model = MLModelSQL(name=name, version=version)
        session.add(model)
        session.commit()

    else:
        model = models[0]

    return model
